from . import grammarette as _grammarette

try:
    from ._version import __version__
except ImportError:
    __version__ = "???"


DEFAULT_DIMS = None


def set_default_dims(dims):
	global DEFAULT_DIMS
	DEFAULT_DIMS = _grammarette._validate_dims(dims)


def induce(lexicon, dims=None):
	if dims is None:
		dims = DEFAULT_DIMS
	return _grammarette.Grammarette(lexicon, dims)



'''
This file was adapted from https://github.com/fbkarsdorp/alignment
'''


from functools import partial
import numpy as np
from scipy.cluster.hierarchy import linkage
from scipy.spatial.distance import squareform


NONE, LEFT, UP, DIAG = 0, 1, 2, 3


def iflatten(iterable):
	"""
	Returns a flattened version of an arbitrarily deeply nested iterable.
	:param iterable: some fixed order iterable (list, tuple)
	:return: generator
	"""
	for elt in iterable:
		if not isinstance(elt, str):
			for sub in flatten(elt):
				yield sub
		else:
			yield elt


def flatten(iterable):
	return list(iflatten(iterable))


def merge(*sequences):
	return list(map(flatten, zip(*sequences)))


def needle_wunsch(sequence_a, sequence_b, scorer, gap_penalty=1, scale=1.0):
	"""
	:param sequence_a: any iterable with a fixed order.
	:param sequence_b: any iterable with a fixed order.
	:param scorer: a dictionary holding the scores between all pairwise
				   items in sequence_a and sequence_b.
	:param gap_penalty: the gap opening penalty used in the analysis.
	:param scale: the factor by which gap_penalty should be decreased.
	:return: numpy matrix, backtrace pointers and the distance between the
			 two sequences.
	"""
	len1, len2 = len(sequence_a), len(sequence_b)
	pointer = np.zeros((len1 + 1, len2 + 1), dtype='i')
	matrix = np.zeros((len1 + 1, len2 + 1), dtype='f')
	length = np.zeros((len1 + 1, len2 + 1), dtype='f')
	pointer[0, 0] = NONE
	pointer[0, 1:] = LEFT
	pointer[1:, 0] = UP
	for i in range(1, len1 + 1):
		matrix[i, 0] = matrix[i - 1, 0] + gap_penalty * scale
		length[i, 0] = length[i - 1, 0] + gap_penalty * scale
	for j in range(1, len2 + 1):
		matrix[0, j] = matrix[0, j - 1] + gap_penalty * scale
		length[0, j] = length[0, j - 1] + gap_penalty * scale
	for i in range(1, len1 + 1):
		for j in range(1, len2 + 1):
			gap_a = matrix[i - 1, j] + (gap_penalty * scale if pointer[i - 1, j] == UP else gap_penalty)
			gap_b = matrix[i, j - 1] + (gap_penalty * scale if pointer[i, j - 1] == LEFT else gap_penalty)
			match = matrix[i - 1, j - 1] + scorer[i - 1, j - 1]
			if gap_a < match and gap_a <= gap_b:
				matrix[i, j] = gap_a
				pointer[i, j] = UP
			elif match <= gap_b:
				matrix[i, j] = match
				pointer[i, j] = DIAG
			else:
				matrix[i, j] = gap_b
				pointer[i, j] = LEFT
			p = pointer[i, j]
			l_gap_a = length[i - 1, j] + (gap_penalty * scale if p == UP else 0)
			l_gap_b = length[i, j - 1] + (gap_penalty * scale if p == LEFT else 0)
			l_match = length[i - 1, j - 1] + (scorer[i - 1, j - 1] if p == DIAG else 0)
			length[i, j] = max(l_gap_a, l_gap_b, l_match)
	# normalize the distance
	distance = matrix[len1, len2] / length[len1, len2]
	return matrix, pointer, distance


def backtrace(pointer, sequence_a, sequence_b):
	i, j = len(sequence_a), len(sequence_b)
	align1, align2 = [], []
	fill_a, fill_b = '_', '_'
	if any(isinstance(e, (tuple, list)) for e in sequence_a):
		fill_a = ('_',) * len(sequence_a[0])
	if any(isinstance(e, (tuple, list)) for e in sequence_b):
		fill_b = ('_',) * len(sequence_b[0])
	while True:
		p = pointer[i, j]
		if p == NONE:
			break
		if p == DIAG:
			align1.append(sequence_a[i - 1])
			align2.append(sequence_b[j - 1])
			i, j = i - 1, j - 1
		elif p == LEFT:
			align1.append(fill_a)
			align2.append(sequence_b[j - 1])
			j -= 1
		elif p == UP:
			align1.append(sequence_a[i - 1])
			align2.append(fill_b)
			i -= 1
		else:
			raise ValueError("Something went terribly wrong.")
	return align1[::-1], align2[::-1]


def align(sequence_a, sequence_b, scores, gap_penalty=1, scale=1.0):
	matrix, pointer, distance = needle_wunsch(
		sequence_a, sequence_b, scores, gap_penalty, scale)
	align1, align2 = backtrace(pointer, sequence_a, sequence_b)
	return align1, align2, distance


def align_sequences(sequence_a, sequence_b, scoring_fn=None, gap_penalty=1, scale=1.0):
	"""
	Align two sequences using the Needleman-Wunsch algorithm.

	:param sequence_a: some fixed order iterable.
	:param sequence_b: some fixed order iterable.
	:param scoring_fn: a distance function.
	:param gap_penalty: the penalty for inserting gaps.
	:param scale: the factor by which gap_penalty should be decreased.
	:return: two new sequences with gaps inserted and the distance between them.
	"""
	if scoring_fn is None:
		scoring_fn = lambda a, b: 0.0 if a == b else 2.0
	scores = {(i, j): scoring_fn(sequence_a[i], sequence_b[j])
			  for i in range(len(sequence_a)) for j in range(len(sequence_b))}
	return align(sequence_a, sequence_b, scores, gap_penalty, scale)


def _align_profiles(sequence_a, sequence_b, scoring_fn=None,
					gap_penalty=1, scale=1.0, gap_weight=0.5):
	scores = {}
	for i in range(len(sequence_a)):
		for j in range(len(sequence_b)):
			dist = 0.0
			count = 0.0
			for k in range(len(sequence_a[i])):
				for l in range(len(sequence_b[j])):
					mi = sequence_a[i][k]
					mj = sequence_b[j][l]
					if mi == '_' or mj == '_':
						dist += gap_weight
					else:
						dist += 0.0 if scoring_fn(mi, mj) < 1 else 1.0
					count += 1.0
				scores[i, j] = dist / count
	return align(sequence_a, sequence_b, scores, gap_penalty, scale)


def pairwise_distances(sequences, fn):
	distances = np.zeros((len(sequences), len(sequences)))
	for i in range(len(sequences)):
		for j in range(i):
			_, _, distance = fn(sequences[i], sequences[j])
			distances[i, j] = distance
			distances[j, i] = distances[i, j]
	return distances


def multi_sequence_alignment(sequences, scoring_fn=None, linking='single',
							 gap_penalty=1, scale=1.0, gap_weight=1.0, verbosity=0):
	"""
	Perform progressive multiple sequence alignment.

	:param sequences: some iterable of fixed order iterables.
	:param scoring_fn: a distance function.
	:param linkage: the linkage function to use for the clustering.
	:return: an Alignment object.
	"""
	if scoring_fn is None:
		scoring_fn = lambda a, b: 0.0 if a == b else 2.0
	# compute all pairwise distances
	matrix = pairwise_distances(sequences, partial(align_sequences, scoring_fn=scoring_fn, gap_penalty=gap_penalty, scale=scale))
	# compute the guiding tree to do the progressive alignment
	Z = linkage(squareform(matrix), method='single')
	# perform the alignment by iterating through the clusters
	alignments = {}
	n_seqs = len(sequences)
	for cluster_id, (node1, node2, _, _) in enumerate(Z, n_seqs):
		node1, node2 = int(node1), int(node2)
		if node1 < n_seqs and node2 < n_seqs:
			align1, align2, _ = align_sequences(sequences[node1], sequences[node2], scoring_fn, gap_penalty, scale)
		else:
			if node1 < n_seqs:
				sequence_a, sequence_b = [[elt] for elt in sequences[node1]], alignments[node2]
			elif node2 < n_seqs:
				sequence_a, sequence_b = alignments[node1], [[elt] for elt in sequences[node2]]
			else:
				sequence_a, sequence_b = alignments[node1], alignments[node2]
			align1, align2, _ = _align_profiles(sequence_a, sequence_b, scoring_fn, gap_penalty, scale, gap_weight)
		alignments[cluster_id] = merge(align1, align2)
	return list(zip(*map(flatten, alignments[max(alignments)])))



from collections import defaultdict
from itertools import product
import grammarette

one_d = {
	(0,): "buviko",
	(0,): "buviko",
	(0,): "buviko",
	(0,): "buviko",
	(1,): "zetiko",
	(1,): "zetiko",
	(1,): "zetiko",
	(1,): "zetiko",
	(2,): "gafiko",
	(2,): "gafiko",
	(2,): "gafiko",
	(2,): "gafiko",
	(3,): "wopiko",
	(3,): "wopiko",
	(3,): "wopiko",
	(3,): "wopiko",
}
transparent = {
	(0, 0): "buviko",
	(0, 1): "buviko",
	(0, 2): "buviko",
	(0, 3): "buviko",
	(1, 0): "zetiko",
	(1, 1): "zetiko",
	(1, 2): "zetiko",
	(1, 3): "zetiko",
	(2, 0): "gafiko",
	(2, 1): "gafiko",
	(2, 2): "gafiko",
	(2, 3): "gafiko",
	(3, 0): "wopiko",
	(3, 1): "wopiko",
	(3, 2): "wopiko",
	(3, 3): "wopiko",
}
holistic = {
	(0, 0): "buvichoe",
	(0, 1): "buvikoh",
	(0, 2): "buvicow",
	(0, 3): "buvycow",
	(1, 0): "zeteekow",
	(1, 1): "zeteecoh",
	(1, 2): "zeticoe",
	(1, 3): "zeticoh",
	(2, 0): "gafeechow",
	(2, 1): "gafeekoe",
	(2, 2): "gafykoe",
	(2, 3): "gafychoe",
	(3, 0): "wopykow",
	(3, 1): "wopeecoe",
	(3, 2): "wopykoh",
	(3, 3): "wopikow",
}
redundant = {
	(0, 0): "buvikoh",
	(0, 1): "buvikoh",
	(0, 2): "buvikoh",
	(0, 3): "buvikoh",
	(1, 0): "zeteeco",
	(1, 1): "zeteeco",
	(1, 2): "zeteeco",
	(1, 3): "zeteeco",
	(2, 0): "gafykoe",
	(2, 1): "gafykoe",
	(2, 2): "gafykoe",
	(2, 3): "gafykoe",
	(3, 0): "wopycow",
	(3, 1): "wopycow",
	(3, 2): "wopycow",
	(3, 3): "wopycow",
}
expressive = {
	(0, 0): "buvikoh",
	(0, 1): "buveeco",
	(0, 2): "buvykoe",
	(0, 3): "buvycow",
	(1, 0): "zetikoh",
	(1, 1): "zeteeco",
	(1, 2): "zetykoe",
	(1, 3): "zetycow",
	(2, 0): "gafikoh",
	(2, 1): "gafeeco",
	(2, 2): "gafykoe",
	(2, 3): "gafycow",
	(3, 0): "wopikoh",
	(3, 1): "wopeeco",
	(3, 2): "wopykoe",
	(3, 3): "wopycow",
}
expressive_letter = {
	(0, 0): "buvikop",
	(0, 1): "buvikog",
	(0, 2): "buvikob",
	(0, 3): "buvikoy",
	(1, 0): "zetikop",
	(1, 1): "zetikog",
	(1, 2): "zetikob",
	(1, 3): "zetikoy",
	(2, 0): "gafikop",
	(2, 1): "gafikog",
	(2, 2): "gafikob",
	(2, 3): "gafikoy",
	(3, 0): "wopikop",
	(3, 1): "wopikog",
	(3, 2): "wopikob",
	(3, 3): "wopikoy",
}
expressive_length = {
	(0, 0): "buviko",
	(0, 1): "buvikoo",
	(0, 2): "buvikooo",
	(0, 3): "buvikoooo",
	(1, 0): "zetiko",
	(1, 1): "zetikoo",
	(1, 2): "zetikooo",
	(1, 3): "zetikoooo",
	(2, 0): "gafiko",
	(2, 1): "gafikoo",
	(2, 2): "gafikooo",
	(2, 3): "gafikoooo",
	(3, 0): "wopiko",
	(3, 1): "wopikoo",
	(3, 2): "wopikooo",
	(3, 3): "wopikoooo",
}
redundant_expressive = {
	(0, 0): "buvikoe",
	(0, 1): "buvikoh",
	(0, 2): "buvikoe",
	(0, 3): "buvichoe",
	(1, 0): "zeteekoe",
	(1, 1): "zeteekoh",
	(1, 2): "zeteekoe",
	(1, 3): "zeteechoe",
	(2, 0): "gafykoe",
	(2, 1): "gafykoh",
	(2, 2): "gafykoe",
	(2, 3): "gaffychoe",
	(3, 0): "wopykoe",
	(3, 1): "wopykoh",
	(3, 2): "wopykoe",
	(3, 3): "wopychoe",
}
kcs_example1 = { # (color, movement, shape)
	(0, 0, 0): "tuge",
	(0, 0, 1): "tuge",
	(0, 0, 2): "tuge",
	(0, 1, 0): "tupim",
	(0, 1, 1): "miniku",
	(0, 1, 2): "tupin",
	(0, 2, 0): "poi",
	(0, 2, 1): "poi",
	(0, 2, 2): "poi",
	(1, 0, 0): "tuge",
	(1, 0, 1): "tuge",
	(1, 0, 2): "tuge",
	(1, 1, 0): "tupim",
	(1, 1, 1): "miniku",
	(1, 1, 2): "tupin",
	(1, 2, 0): "poi",
	(1, 2, 1): "poi",
	(1, 2, 2): "poi",
	(2, 0, 0): "tuge",
	(2, 0, 1): "tuge",
	(2, 0, 2): "tuge",
	(2, 1, 0): "tupim",
	(2, 1, 1): "miniku",
	(2, 1, 2): "tupin",
	(2, 2, 0): "poi",
	(2, 2, 1): "poi",
	(2, 2, 2): "poi",
}
kcs_example2 = { # (color, movement, shape)
	(0, 0, 0): "nereki",
	(0, 0, 1): "neheki",
	(0, 0, 2): "nekeki",
	(0, 1, 0): "nereplo",
	(0, 1, 1): "nehoplo",
	(0, 1, 2): "nekiplo",
	(0, 2, 0): "nepilu",
	(0, 2, 1): "nehopilu",
	(0, 2, 2): "nekipilu",
	(1, 0, 0): "lereki",
	(1, 0, 1): "lahoki",
	(1, 0, 2): "lakeki",
	(1, 1, 0): "laneplo",
	(1, 1, 1): "lahoplo",
	(1, 1, 2): "lakiplo",
	(1, 2, 0): "lanepilu",
	(1, 2, 1): "lahopilu",
	(1, 2, 2): "lakipilu",
	(2, 0, 0): "renana",
	(2, 0, 1): "reneki",
	(2, 0, 2): "raheki",
	(2, 1, 0): "replo",
	(2, 1, 1): "rehoplo",
	(2, 1, 2): "rahoplo",
	(2, 2, 0): "repilu",
	(2, 2, 1): "rehopilu",
	(2, 2, 2): "rahopilu",
}
ktcs_example1 = {
	(0, 0): "pihino",
	(0, 1): "nemone",
	(0, 2): "piga",
	(0, 3): "kawake",
	(1, 0): "kapa",
	(1, 1): "gakho",
	(1, 2): "wuwele",
	(1, 3): "nepi",
	(2, 0): "newhomo",
	(2, 1): "kamone",
	(2, 2): "gaku",
	(2, 3): "hokako",
}
ktcs_example2 = {
	(0, 0): "egewawu",
	(0, 1): "egewawa",
	(0, 2): "egewuwu",
	(0, 3): "ege",
	(1, 0): "mega",
	(1, 1): "megawawa",
	(1, 2): "megawuwu",
	(1, 3): "wulagi",
	(2, 0): "gamanewawu",
	(2, 1): "gamanewawa",
	(2, 2): "gamanewuwu",
	(2, 3): "gamane",
}

test_set = [
	(one_d, (4,), '0buv|1zet|2gaf|3wop+?iko'),
	(transparent, (4, 4), '0?buv|1?zet|2?gaf|3?wop+??iko'),
	(holistic, (4, 4), '0?buv|1?zet|2?gaf|3?wop+00ichoe|01ikoh|02icow|03ycow|10eekow|11eecoh|12icoe|13icoh|20eechow|21eekoe|22ykoe|23ychoe|30ykow|31eecoe|32ykoh|33ikow'),
	(redundant, (4, 4), '0?buvikoh|1?zeteeco|2?gafykoe|3?wopycow'),
	(expressive, (4, 4), '0?buv|1?zet|2?gaf|3?wop+?0ikoh|?1eeco|?2ykoe|?3ycow'),
	(expressive_letter, (4, 4), '0?buv|1?zet|2?gaf|3?wop+??iko+?0p|?1g|?2b|?3y'),
	(expressive_length, (4, 4), '0?buv|1?zet|2?gaf|3?wop+??iko+?1o|?2oo|?3ooo'),
	(redundant_expressive, (4, 4), '0?buvi|1?zetee|23gaffy|2?gafy|3?wopy+?3ch|??k+?1oh|??oe'),
	(kcs_example1, (3, 3, 3), '?0?tuge|?10tupim|?11miniku|?12tupin|?2?poi'),
	(kcs_example2, (3, 3, 3), '001nehe|002neke|020ne|0?0nere|0?1neho|0?2neki|100lere|102lake|1?0lane|1?1laho|1?2laki|200ren|201rene|202rahe|2?0re|2?1reho|2?2raho+200ana|?0?ki|?1?plo|?2?pilu'),
	(ktcs_example1, (3, 4), '00pihino|01nemone|02piga|03kawake|10kapa|11gakho|12wuwele|13nepi|20newhomo|21kamone|22gaku|23hokako'),
	(ktcs_example2, (3, 4), '0?ege|10meg|13wulagi|1?mega|2?gamane+10a|?0wawu|?1wawa|?2wuwu'),
]

def test_grammar(grammar, lexicon):
	signal_to_meanings = defaultdict(set)
	for meaning, expected_signal in lexicon.items():
		signal_to_meanings[expected_signal].add(meaning)
		produced_signal = grammar.produce(meaning)
		if expected_signal != produced_signal:
			print(f'❗️ Production error for {meaning}: Expected "{expected_signal}", produced "{produced_signal}"')
	for signal, expected_meanings in signal_to_meanings.items():
		inferred_meanings = grammar.comprehend(signal)
		if inferred_meanings != expected_meanings:
			print(f'❗️ Comprehension error for "{signal}": Expected {expected_meanings}, inferred {inferred_meanings}')

for test_lexicon, dims, expected_grammar in test_set:

	G = grammarette.induce(test_lexicon, dims)
	print(G)
	print(G.codelength)

	assert G.grammar == expected_grammar
	test_grammar(G, test_lexicon)



import re
from collections import defaultdict
from itertools import product
import numpy as np
from .msa import multi_sequence_alignment


VOWELS = ['a','e','i','o','u','y']
CONSONANTS = ['b','c','d','f','g','h','j','k','l','m','n','p','q','r','s','t','v','w','x','z']


class Grammarette:

	def __init__(self, lexicon, dims):
		self.dims = _validate_dims(dims)
		self.n_features = len(self.dims)
		self.lexicon = _validate_lexicon(lexicon, self.dims)
		self.signals = [signal for meaning, signal in lexicon.items()]
		self.meanings = [tuple(map(str, meaning)) for meaning, signal in lexicon.items()]
		self.lexicon_reverse = defaultdict(set)
		for meaning, signal in zip(self.meanings, self.signals):
			self.lexicon_reverse[signal].add(meaning)
		self.shortest_grammar = '|'.join([
			''.join(meaning) + signal for meaning, signal in zip(self.meanings, self.signals)
		])
		self.shortest_codelength = self._codelength(self.shortest_grammar)
		self.grammars = {self.shortest_grammar: self.shortest_codelength}
		self.alignment = self._align_signals()
		self.alignment_reverse = self._align_signals(reverse=True)
		self._find_shortest_grammar()
		self._regex = None
		self._regexc = None

	def __str__(self):
		return f'Grammarette[{self.shortest_grammar}]'

	def _align_signals(self, reverse=False):
		'''
		Produce an alignment of the signals. If reverse is True, the signals
		are first reversed and later unreversed to produce an alignment
		working from the end backwards.
		'''
		if reverse:
			sequences = [list(reversed(signal)) for signal in self.signals]
		else:
			sequences = [list(signal) for signal in self.signals]
		alignment = multi_sequence_alignment(sequences, scoring_fn=_letter_distance)
		# Since multi_sequence_alignment() doesn't retain the original
		# sequence order, we now need to put them back into order.
		signal_checklist = [''.join(sequence) for sequence in sequences]
		ordered_alignment = [None] * len(alignment)
		for aligned_signal in alignment:
			signal_i = signal_checklist.index(''.join(aligned_signal).replace('_', ''))
			signal_checklist[signal_i] = None
			if reverse:
				ordered_alignment[signal_i] = tuple(reversed(aligned_signal))
			else:
				ordered_alignment[signal_i] = aligned_signal
		return np.array(ordered_alignment, dtype=str)

	def _codelength(self, grammar):
		'''
		Calculate the codelength of a grammar string. This is given by

		  -sum( log[Pr(char)] for char in grammar )

		where Pr(char) is the probability of some character appearing in the
		grammar.
		'''
		unique_chars = set(grammar)
		grammar_length = len(grammar)
		char_codelengths = {char: np.log2(grammar.count(char) / grammar_length) for char in unique_chars}
		return -sum([char_codelengths[char] for char in grammar])

	def _produce_signal_for_meaning(self, grammar, target_meaning):
		'''
		Produce a signal for a target meaning according to a given grammar.
		'''
		signal = ''
		for rule in grammar.split('+'):
			for sub_rule in rule.split('|'):
				meaning = tuple(sub_rule[:self.n_features])
				form = sub_rule[self.n_features:]
				if all([m_val == t_val or m_val == '?' for m_val, t_val in zip(meaning, target_meaning)]):
					signal += form
					break
		return signal

	# def _infer_meanings_for_signal(self, grammar, target_signal):
	# 	# grammar = '00buvichoe|01buvikoh|02buvicow|03buvycow|10zeteekow|11zeteecoh|12zeticoe|13zeticoh|20gafeechow|21gafeekoe|22gafykoe|23gafychoe|30wopykow|31wopeecoe|32wopykoh|33wopikow'
	# 	# target_signal = 'buvikoh'
	# 	candidate_meanings = set(product(*[map(str, range(n_values)) for n_values in self.dims]))
	# 	for rule in grammar.split('+'):
	# 		for sub_rule in rule.split('|'):
	# 			meaning = sub_rule[:self.n_features]
	# 			form = sub_rule[self.n_features:]
	# 			if target_signal.startswith(form):
	# 				target_signal = target_signal[len(form):]
	# 				if '?' not in meaning:
	# 					candidate_meanings.add(tuple(meaning))
	# 				for feature_i, value in enumerate(meaning):
	# 					if value == '?':
	# 						continue
	# 					candidate_meanings = set(filter(lambda m: True if m[feature_i] == value else False, candidate_meanings))
	# 				break
	# 			else:
	# 				for feature_i, value in enumerate(meaning):
	# 					if value == '?':
	# 						continue
	# 					candidate_meanings = set(filter(lambda m: False if m[feature_i] == value else True, candidate_meanings))
	# 	# print(candidate_meanings)
	# 	# quit()
	# 	return candidate_meanings
	# 	return self.lexicon_reverse[target_signal]

	def _check_consistency(self, grammar):
		'''
		For each meaning–signal pair in the original input lexicon, check that
		the grammar produces the correct output signal.
		'''
		for meaning, expected_signal in zip(self.meanings, self.signals):
			produced_signal = self._produce_signal_for_meaning(grammar, meaning)
			if expected_signal != produced_signal:
				return False
		# for signal, expected_meanings in self.lexicon_reverse.items():
		# 	inferred_meanings = self._infer_meanings_for_signal(grammar, signal)
		# 	if set(inferred_meanings) != expected_meanings:
		# 		return False
		return True

	def _evaluate_grammar(self, grammar):
		'''
		If the grammar hasn't been observed before and is consistent with the
		input data, calculate its codelength, store it in self.grammars,
		and keep track of the shortest grammar observed so far.
		'''
		if grammar not in self.grammars:
			if self._check_consistency(grammar):
				codelength = self._codelength(grammar)
				self.grammars[grammar] = codelength
				if codelength < self.shortest_codelength:
					self.shortest_grammar = grammar
					self.shortest_codelength = codelength

	def _intersection(self, meanings):
		'''
		Given a list of meanings, return the intersection. If the meanings
		differ on only one feature, change that feature to the wildcard ?.
		For example, given [(2, 2), (2, 1)], return (2, ?).
		'''
		different_features = []
		for feature_i, values in enumerate(zip(*meanings)):
			if len(set(values)) > 1:
				different_features.append(feature_i)
		if len(different_features) == 1:
			intersection = list(meanings[0])
			intersection[different_features[0]] = '?'
			return tuple(intersection)
		return False

	def _reduce_meanings(self, chunk_meanings):
		'''
		Given a list of meanings, if, for feature i, all feature values are
		represented, reduce those feature values to the wildcard '?'. For
		example, given [(0, 1), (0, 2), (0, 3), (0, 4)], return [(0, ?)].
		'''
		reduction = []
		for feature_i, n_values in enumerate(self.dims):
			values_for_feature_i = set([meaning[feature_i] for meaning in chunk_meanings])
			if len(values_for_feature_i) == n_values:
				values_for_feature_i = ['?']
			reduction.append(values_for_feature_i)
		reduced_meanings = []
		for meaning in product(*reduction):
			if '?' in meaning or meaning in chunk_meanings:
				reduced_meanings.append(meaning)
		return reduced_meanings

	def _write_grammar(self, alignment, chunk_boundaries, reduce_sub_grammar=False):
		'''
		Given an alignment of the signals and boundaries for carving up the
		alignment, write a grammar. Within each chunk boundary, get a list
		of the chunks that occur. For each unique chunk, make a list of
		all the meanings it occurs alongside (and reduce that set of
		meanings, if possible). If reduce_sub_grammar is True, in cases
		where a particular unique chunk maps to multiple meanings, try to
		reduce those meanings further by taking their intersection.
		Finally, join the grammar into a string and evaluate it.
		'''
		grammar = []
		for start, end in chunk_boundaries:
			chunks = [''.join(chunk).replace('_', '') for chunk in alignment[:, start:end]]
			chunk_to_meanings = {}
			for unique_chunk in set(chunks):
				if len(unique_chunk) == 0:
					continue
				chunk_to_meanings[unique_chunk] = self._reduce_meanings(
					[self.meanings[i] for i, chunk in enumerate(chunks) if chunk == unique_chunk]
				)
			sub_grammar = []
			for chunk, meanings in chunk_to_meanings.items():
				if reduce_sub_grammar and len(meanings) > 1:
					if (intersection := self._intersection(meanings)) is not False:
						sub_grammar.append(''.join(intersection) + chunk)
						continue
				for meaning in meanings:
					sub_grammar.append(''.join(meaning) + chunk)
			# sub_grammar.sort()
			# grammar.append('|'.join(sorted(sub_grammar, key=lambda x: len(x), reverse=True)))
			grammar.append('|'.join(sorted(sub_grammar)))
		grammar = '+'.join(grammar)
		self._evaluate_grammar(grammar)

	def _find_shortest_grammar(self):
		'''
		For each of the two alignments (forward and reverse), iterate over all
		possible chunk boundaries (partitions of the alignment columns)
		and write a grammar for each chunking of the lexicon.
		'''
		for alignment in [self.alignment, self.alignment_reverse]:
			for chunk_boundaries in _iter_partition_boundaries(len(alignment[0])):
				self._write_grammar(alignment, chunk_boundaries, reduce_sub_grammar=False)
				self._write_grammar(alignment, chunk_boundaries, reduce_sub_grammar=True)

	@property
	def grammar(self):
		return self.shortest_grammar

	@property
	def codelength(self):
		return self.shortest_codelength

	@property
	def regex(self):
		'''
		Return the induced grammar encoded as a regular expression. Meanings
		are encoded in the named groups (with _ representing the wildcard).
		'''
		if self._regex is None:
			self._regex = '^'
			for rule in self.shortest_grammar.split('+'):
				sub_rule_strings = []
				for sub_rule in rule.split('|'):
					meaning = sub_rule[:self.n_features]
					form = sub_rule[self.n_features:]
					meaning_id = f'{generate_random_group_name()}{meaning.replace("?", "_")}'
					sub_rule_strings.append(f'(?P<{meaning_id}>{form})')
				self._regex += '(' + '|'.join(sub_rule_strings) + ')?'
			self._regex += '$'
		return self._regex

	@property
	def regexc(self):
		'''
		Return the compiled regular expression.
		'''
		if self._regexc is None:
			self._regexc = re.compile(self.regex)
		return self._regexc

	def produce(self, meaning):
		'''
		Given a meaning, produce a signal according to the induced grammar.
		'''
		return self._produce_signal_for_meaning(self.grammar, tuple(map(str, meaning)))

	# def comprehend(self, signal):
	# 	return set([tuple(map(int, meaning)) for meaning in self._infer_meanings_for_signal(self.grammar, signal)])

	def comprehend(self, signal):
		'''
		Given a signal, return the possible meanings according to the induced
		grammar. To achieve this we first convert the induced grammar to a
		regular expression, match it against the signal, and infer the
		possible meanings from the named groups required to parse the signal.
		'''
		inferred_meaning = ['?'] * self.n_features
		for meaning_id, form in self.regexc.match(signal).groupdict().items():
			if form is not None:
				meaning_id = meaning_id[-self.n_features:]
				for feature_i, value in enumerate(meaning_id):
					if value != '_':
						inferred_meaning[feature_i] = value
		candidate_meanings = []
		for meaning in self.meanings:
			if all([m_val == i_val or i_val == '?' for m_val, i_val in zip(meaning, inferred_meaning)]):
				candidate_meanings.append(tuple(map(int, meaning)))
		return set(candidate_meanings)


def _letter_distance(a, b):
	'''
	Return the distance between two characters. If the characters are
	identical, return 0. If both characters belong to the same category
	(vowels or consonants), return 1. Otherwise, return 2.
	'''
	if a == b:
		return 0.0
	if a in VOWELS and b in VOWELS:
		return 1.0
	if a in CONSONANTS and b in CONSONANTS:
		return 1.0
	return 2.0


def _iter_partition_boundaries(n):
	'''
	Yield the boundaries needed to partition an object of size n in all ways.
	For example, an object of length 3, like ABC, can be partitioned in
	2^(n-1) = 4 ways: (ABC), (AB C), (A BC),(A B C).
	'''
	for cutpoints in range(1 << (n - 1)):
		result = []
		lastcut = 0
		for i in range(n - 1):
			if (1 << i) & cutpoints != 0:
				result.append((lastcut, i + 1))
				lastcut = i+1
		result.append((lastcut, n))
		yield tuple(result)


def _validate_lexicon(lexicon, dims):
	if not isinstance(lexicon, dict):
		raise TypeError(f'Lexicon should be of type dict, got {lexicon.__class__.__name__}')
	for meaning, signal in lexicon.items():
		if not isinstance(meaning, tuple):
			raise TypeError(f'Meanings should be of type tuple, got {meaning.__class__.__name__}')
		if not isinstance(signal, str):
			raise TypeError(f'Signals should be of type str, got {signal.__class__.__name__}')
		if len(meaning) != len(dims):
			raise ValueError(f'Meaning {meaning} does not match the dimensionality {dims}')
		for feature_value, n_features in zip(meaning, dims):
			if feature_value >= n_features:
				raise ValueError(f'Meaning {meaning} is not within the dimensionality {dims}')
	return lexicon


def _validate_dims(dims):
	if not isinstance(dims, tuple):
		raise TypeError(f'Dims should be of type tuple, got {dims.__class__.__name__}')
	if len(dims) == 0:
		raise ValueError('Dims should have at least one dimension')
	for n_features in dims:
		if not isinstance(n_features, int):
			raise TypeError(f'Dims should be tuple of ints, got tuple containing {n_features.__class__.__name__}')
	return dims


def generate_random_group_name():
	name = ''
	for _ in range(4):
		name += np.random.choice(CONSONANTS)
		name += np.random.choice(VOWELS)
	return name



import setuptools

with open("README.md", encoding="utf-8") as file:
    long_description = file.read()

setuptools.setup(
    name="grammarette",
    use_scm_version={
        "write_to": "grammarette/_version.py",
        "write_to_template": '__version__ = "{version}"',
        "fallback_version": "???",
    },
    author="Jon Carr",
    author_email="jon.carr@rhul.ac.uk",
    description="A lightweight grammar inducer for toy lexicons",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/jwcarr/grammarette",
    license="GPLv3",
    packages=["grammarette"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
    ],
    python_requires=">=3.8",
    install_requires=["numpy>=1.20", "scipy>=1.5"],
    setup_requires=["setuptools_scm"],
)
